var template = require('babel-template');

var mixinTpl = template(`
    MIXINS && _ko_component_mixin(this, opts, MIXINS);
`);

var computedTpl = template(`
    this.NAME = ko.computed(VALUE, this);
`);

var pureComputedTpl = template(`
    this.NAME = ko.pureComputed(VALUE, this);
`);

var componentTpl = template(`
    METHODS && Object.assign(constructor.prototype, METHODS);

    ko.components.register(NAME, {
        synchronous: true,
        viewModel: {
            defaults: DEFAULTS,
            createViewModel: function (params, componentInfo) {
                return new constructor(params, componentInfo);
            }
        },
        style: STYLE,
        template: TEMPLATE,
    });
`);

var nameKey = '_ko_component_name';
var styleKey = '_ko_component_style';
var templateKey = '_ko_component_template';
var computedList = [];

module.exports = function (babel) {
    var t = babel.types;

    var ast = null;
    var constructorBlockPath = null;

    // Normalize identifier
    var normalize = function (str) {
        var words = null;

        function upperCamel (key) {
            return key.charAt(0).toUpperCase() + key.substr(1);
        };

        if (str) {
            words = str.split(/[\-\_]/g);
            return words.map(upperCamel).join('');
        } else {
            return '';
        }
    };

    // Reset component partial ast
    var clean = function () {
        ast = Object.create(null);
        constructorBlockPath = null;
    };

    // Render component template into ast
    var render = function () {
        return componentTpl({
            NAME: ast[nameKey],
            CTOR_NAME: t.Identifier(normalize(ast[nameKey].value)),
            DEFAULTS: ast['defaults'] || t.nullLiteral(),
            METHODS: ast['methods'] || t.nullLiteral(),
            STYLE: ast[styleKey] || t.stringLiteral(''),
            TEMPLATE: ast[templateKey] || t.stringLiteral('')
        });
    };

    // Detect whether if a valid component module
    var validateComponent = function () {
        if (!t.isStringLiteral(ast[nameKey])) {
            return false;
        }

        if (ast[nameKey] === '') {
            throw new Error('Component name is empty string.');
        }

        if (!ast['constructor'] || t.isNullLiteral(ast['constructor'])) {
            throw new Error('Can not read "constructor". Does this exported?');
        }

        return true;
    };

    var constructorVisitor = {
        FunctionExpression: {
            enter: function (path) {
                path.node.id = t.identifier(normalize(ast[nameKey].value));
            }
        },
        BlockStatement: {
            enter: function (path) {
                constructorBlockPath = path;
            }
        }
    };

    var computedVisitor = {
        Property: {
            enter: function (path) {
                computedList.push(computedTpl({
                    NAME: path.key,
                    VALUE: path.value
                }));
            }
        }
    };

    var pureComputedVisitor = {
        Property: {
            enter: function (path) {
                computedList.push(pureComputedTpl({
                    NAME: path.key,
                    VALUE: path.value
                }));
            }
        }
    };

    var variableVisitor = {
        VariableDeclarator: {
            enter: function (path) {
                var grandPathType = path.parentPath.parentPath.type;
                var declarationName = path.node.id.name;

                if (grandPathType !== 'Program' && grandPathType !== 'ExportNamedDeclaration') {
                    return;
                }

                switch (declarationName) {
                    case nameKey:
                        ast[declarationName] = path.node.init;
                        break;
                    case 'constructor':
                        ast[declarationName] = path.node.init;
                        path.traverse(constructorVisitor);
                        break;
                    case 'computed':
                        path.traverse(computedVisitor);
                        break;
                    case 'pureComputed':
                        path.traverse(pureComputedVisitor);
                        break;
                    default:
                        ast[declarationName] = path.node.id;
                        break;
                }
            }
        }
    };

    clean();

    return {
        visitor: {
            VariableDeclarator: variableVisitor.VariableDeclarator,

            ExportNamedDeclaration: {
                enter: function (path) {
                    path.traverse(variableVisitor);
                }
            },

            Program: {
                exit: function (path) {
                    var declaration = null;

                    if (validateComponent()) {
                        path.pushContainer('body', render());

                        if (constructorBlockPath) {
                            constructorBlockPath.pushContainer('body', mixinTpl({
                                MIXINS: ast['mixins'] || t.nullLiteral()
                            }));

                            computedList.forEach(function (computed) {
                                constructorBlockPath.pushContainer('body', computed);
                            });
                        }
                    }

                    clean();
                }
            }
        }
    };
};
